/**
IBM SOFTWARE DISCLAIMER 

Java array package (draft 0.2). Copyright (1998), International Business 
Machines Corporation.

Permission to use, copy, modify and distribute this software for any
noncommercial purpose and without fee is hereby granted, provided that
this copyright and permission notice appear on all copies of the
software. The name of the IBM Corporation may not be used in any
advertising or publicity pertaining to the use of the software. IBM
makes no warranty or representations about the suitability of the
software for any purpose.  It is provided "AS IS" without any express
or implied warranty, including the implied warranties of
merchantability, fitness for a particular purpose and non-infringement.
IBM shall not be liable for any direct, indirect, special or
consequential damages resulting from the loss of use, data or projects,
whether in an action of contract or tort, arising out of or in
connection with the use or performance of this software.
*/

package array;

/**
   Blas = Basic Linear Algebra Subprograms for the array package.
   <P>
   This class implements the BLAS functions for operations on
   arrays from the array package.
   It follows the spirit of the 
   "Draft Proposal for Java BLAS Interface", by Roldan Pozo of the
   National Institute of Standards and Technology.
   Because the array package supports sections, the interface is 
   actually simpler.
   <P>
   The purpose of this current implementation is to serve as an example
   of interfaces. Therefore, only three operations are coded: 
   <b>dgemm</b>, <b>dgemv</b>, and <b>ddot</b>.
   Please note that the operations are coded for clarity, not performance.
*/

public class Blas {

    public static final boolean NoTranspose = false;
    public static final boolean Transpose = true;

    private static int min(int a, int b) {

	return (a < b) ? a : b;
    }
    /**
       BLAS dgemm routine: C = alpha*A*B + beta*C.
       @param transa	indicates if A is to be transposed
       @param transb    indicates if B is to be transposed
       @param alpha     multiplier for A*B
       @param a         matrix A
       @param b		matrix B
       @param beta	multiplier for C
       @param c		matrix C
       @exception NonconformingArrayException shapes do not match
    */
    public static void dgemm(boolean transa,
			     boolean transb,
			     double alpha,
			     doubleArray2D a,
			     doubleArray2D b,
			     double beta,
			     doubleArray2D c)
	throws NonconformingArrayException {

	try {
	    if (transa) a = a.permuteAxes(1,0);
	    if (transb) b = b.permuteAxes(1,0);

	    int m = a.size(0);
	    int n = a.size(1);
	    int p = b.size(1);

	    if (n != b.size(0)) throw new NonconformingArrayException();
	    if (p != c.size(1)) throw new NonconformingArrayException();
	    if (m != c.size(0)) throw new NonconformingArrayException();

	    doubleArray2D d = new doubleArray2D(m,p);

	    for (int i=0; i<m; i++) {
		for (int j=0; j<p; j++) {
		    double s = 0;
		    for (int k=0; k<n; k++) {
			s += a.getf(i,k)*b.getf(k,j);
		    }
		    d.setf(i,j,alpha*s+beta*c.getf(i,j));
		}
	    }
	    c.assign(d);
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in Blas.dgemm()");
	    System.exit(-1);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in Blas.dgemm()");
	    System.exit(-1);
	}
    }

    /**
       BLAS ddot routine: s = x^T * y.
       @param x		vector x
       @param y		vector y
       @exception NonconformingArrayException x and y must be of same length
    */
    public static double ddot(doubleArray1D x,
			      doubleArray1D y)
	throws NonconformingArrayException {

	double s = 0;
	try {
	    int n = x.size(0);

	    if (n != y.size(0)) throw new NonconformingArrayException();

	    for (int i=0; i<n; i++) {
		s += x.getf(i)*y.getf(i);
	    }
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in Blas.ddot()");
	    System.exit(-1);
	}

	return s;
    }

    /**
       BLAS dgemv routine: y = alpha*A*x + beta*y.
       @param transa	indicates if A is to be transposed
       @param alpha	multiplier for A*x
       @param a		matrix A
       @param x		vector x
       @param beta	multiplier for y
       @param y		vector y
       @exception NonconformingArrayException shapes do not match
    */
    public static void dgemv(boolean transa,
			     double alpha,
			     doubleArray2D a,
			     doubleArray1D x,
			     double beta,
			     doubleArray1D y) 
	throws NonconformingArrayException {

	try {
	    if (transa) a = a.permuteAxes(1,0);

	    int m = a.size(0);
	    int n = a.size(1);

	    if (n != x.size(0)) throw new NonconformingArrayException();
	    if (m != y.size(0)) throw new NonconformingArrayException();

	    doubleArray1D z = new doubleArray1D(m);

	    for (int i=0; i<m; i++) {
		double s = 0;
		for (int k=0; k<n; k++) {
		    s += a.getf(i,k)*x.getf(k);
		}
		z.setf(i,alpha*s+beta*y.getf(i));
	    }
	    y.assign(z);
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Internal error in Blas.dgemv()");
	    System.exit(-1);
	} catch (InvalidArrayShapeException e) {
	    System.out.println("Internal error in Blas.dgemv()");
	    System.exit(-1);
	}
    }
}
